/***********************************************
 *  buff.c 
 *                                            
 *  Copyright 2001 Eric Kratzer.               
 *
 *  This file is part of imp3sh.              
 *  Includes Xaudio software Copyright (C)
 *           1996-2000 Xaudio Corporation
 *           All Rights Reserved.
 *                                            
 *  imp3sh is free software; you can distribute
 *  it and/or modify it under the terms of the
 *  GNU General Public License as published by
 *  the Free Software Foundation; either version
 *  2 of the License, or (at your option) any
 *  later version.
 *
 *  imp3sh is distributed in the hope that
 *  it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *  PURPOSE. See the GNU General Public License
 *  for more details.
 *
 *  You should have recieved a copy of the GNU
 *  General Public License along with imp3sh; if
 *  not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston MA
 *  02111-1307 USA
 ************************************************/

/************************************
 *          Include Files           *
 ************************************/

#include "imp3sh.h"
#include "all.h"

/************************************
 *       External References        *
 ************************************/

/* command.c */
extern CommandDef com[];
extern DCharBuf *args;
extern flag_t prompt_now, ignore_char_input;

/* main.c */
extern volatile int tty_height, tty_width, tab_length;
extern flag_t piped_io_flag;

/* shell.c */
extern flag_t ignore_hist_input;

/************************************
 *        Global Variables          *
 ************************************/

DCharBuf *dlog;

/************************************
 *         Local Variables          *
 ************************************/

static DCharBuf *outb, *conb; 
static flag_t mess_search;
static int mess_patlen;
static char mess_msg[INFO_SIZE], mess_pat[INFO_SIZE];
static char tmpstr_bf[BUFF_SIZE];

/************************************
 *         Module routines          * 
 ************************************/

unsigned int buff_get_bytec()
{
    unsigned int bytec;

    bytec = 0;
    bytec += dbuf_get_bytec(outb);
    bytec += dbuf_get_bytec(conb);
    bytec += dbuf_get_bytec(dlog);

    return bytec;
}

void buff_init()
{
    mess_patlen = mess_search = 0;
    memset(mess_msg,  0, INFO_SIZE);
    memset(mess_pat,  0, INFO_SIZE);
    memset(tmpstr_bf, 0, BUFF_SIZE);

    outb = dbuf_init("Mess buffer", DCHAR_GROW, CONB_SIZE, CONB_SIZE*2);
    conb = dbuf_init("Console buffer", DCHAR_SHIFT, CONB_SIZE, 0);
    dlog = dbuf_init("Debug log", DCHAR_SHIFT, DLOG_SIZE, 0);
}

void buff_clear()
{
    dbuf_free(outb);
    dbuf_free(conb);
    dbuf_free(dlog);
}

/************************************
 *     Buffered Output Functions    *
 ************************************/

void outb_sort()
{
    dbuf_sort_str(outb);
}

int outb_empty()
{
    return (outb->pos == outb->len) ? TRUE : FALSE;
}

void outb_pat_add(char in)
{
    if (mess_patlen < (INFO_SIZE-1))
	mess_pat[mess_patlen++] = in;
}

void outb_pat_remove()
{
    if (mess_patlen > 0)
	mess_patlen --;
}

void outb_pat_disp()
{
    mess_pat[mess_patlen] = 0;
    out_str_n("/"); 
    out_str_n(mess_pat);
    out_flush();
}

void outb_pat_search()
{
    int i;
    DCBNode *buf;

    buf = outb->buf;
    mess_pat[mess_patlen] = 0;
    
    for(i = outb->pos; i < outb->len; i++)
	if (grep_str(mess_pat, buf[i].ptr))
	    break;
    
    if (i != outb->len) {
	outb->pos = i;
	outb->var = outb->pos + (tty_height-1);
	out_str("...skipping");
	outb_flush();
    } else {
	out_str_n("Pattern not found");
	out_flush();
    }
}

void outb_clear()
{
    dbuf_clear(outb);
    outb->var = tty_height;
    mess_patlen = *mess_pat = 0;

    /* keep output buffer size decent... */
    if (outb->size > (CONB_SIZE * 8))
	dbuf_resize(outb, CONB_SIZE * 6);
}

void outb_add(const char *frmt, ...)
{
    int len;
    va_list ap;

    va_start(ap, frmt);
    len = vsnprintf(tmpstr_bf, BUFF_SIZE-1, frmt, ap);
    va_end(ap);

    if (len == -1)
        ;
    else if (piped_io_flag)
	out_str(tmpstr_bf);
    else
        dbuf_add(outb, tmpstr_bf, ++len, 0);
}

void outb_msg_disp()
{
    float perc;
    
    perc = 0.0;
    if (outb->len)
	perc = 100.0 * ((float) outb->pos / (float) outb->len);

    com_ctrl_str(1, TC_CR);
    com_ctrl_str(1, TC_CLR_TOEOL);

    safe_sprintf(mess_msg, INFO_SIZE, "--Mess--(%d%%)--", (int) perc);
    out_str_n(mess_msg);
    out_flush();
}

void outb_flush()
{
    int len, cut, tabl;
    DCBNode *buf;

    if (ignore_char_input) {
	conb_add(DEBUG_MSG " %s flush delayed...", outb->name);
	return;
    }

    buf = outb->buf;

    while (outb->pos < outb->len) {	
	len = tabl = 0;
	strncpy(tmpstr_bf, buf[outb->pos].ptr, tty_width + 2);
	tmpstr_bf[tty_width+1] = ZCHAR;

	while (tmpstr_bf[len] != ZCHAR)
	    if (tmpstr_bf[len++] == '\t')
		tabl += tab_length - len%tab_length;

	if ((len+tabl) > tty_width) {
	    cut = tty_width - tabl;
	    if (cut > 3) safe_strcpy(&tmpstr_bf[cut - 3], "...\0", 4);
	}

	if ((outb->pos >= (outb->var - 1))) {
	    outb_msg_disp();
	    break;
	} else {
	    out_str(tmpstr_bf);
	    outb->pos++;
	}
    }

    if (outb_empty()) {
        outb_clear();
	conb_flush();
	if (!prompt_now)
	    update_prompt(FALSE);
    } else
	prompt_now = FALSE;
}

void outb_pat_check(char in)
{
    switch (in) {
    case '\n':
	mess_search = FALSE;
	if (mess_patlen)
	    outb_pat_search();
	else
	    outb_msg_disp();
	break;
    case 8:
    case 127:
	outb_pat_remove();
	outb_pat_disp();
	break;
    default:
	outb_pat_add(in);
	outb_pat_disp();
	break;
    }
}

void outb_ctrl_check(char in)
{
    switch (in) {
    case '/':
	mess_search = TRUE;
	mess_patlen = *mess_pat = 0;
	outb_pat_disp();
	return;
    case 'n':
    case 'N':
	if (mess_patlen)
	    outb_pat_search();
	else {
	    out_str_n("No previous pattern");
	    out_flush();
	}
	return;
    case 'q':
    case 'Q':
	outb_clear();
	conb_flush();
	update_prompt(FALSE);
	return;
	
    case '\n':
	outb->var++;
	break;
    case 'y':
    case 'Y':
	outb->pos -= tty_height;
	outb->var --;
	
	if (outb->pos < 0)
	    outb->pos = 0;
	if (outb->var < tty_height)
	    outb->var = tty_height;
	break;
    case 'b':
    case 'B':
	outb->pos -= (tty_height - 1) * 2;
	outb->var -= (tty_height - 1);
	
	if (outb->pos < 0)
	    outb->pos = 0;
	if (outb->var < tty_height)
	    outb->var = tty_height;
	break;
    default:
	outb->var += (tty_height - 1);
	break;
    }
    
    outb_flush();
}

int outb_check(char in)
{
    if (outb_empty())
	return FALSE;

    com_ctrl_str(1, TC_CR);
    com_ctrl_str(1, TC_CLR_TOEOL);

    if (mess_search)
	outb_pat_check(in);
    else
	outb_ctrl_check(in);

    return TRUE;
}

/************************************
 *      Console Buffer Functions    *
 ************************************/

void conb_flush()
{
    int i;
    DCBNode *buf;

    if (!have_tty() || !conb->len)
	return;

    buf = conb->buf;

    for (i = 0; i < conb->len; i++)
	if (buf[i].info == CONB_NO_NEWLINE)
	    out_str_n(buf[i].ptr);
	else
	    out_str(buf[i].ptr);
    
    dbuf_clear(conb);
}

void conb_add(const char *frmt, ...)
{
    int len;
    va_list ap;

    va_start(ap, frmt);
    len = vsnprintf(tmpstr_bf, BUFF_SIZE-1, frmt, ap);
    va_end(ap);

    if (len == -1)
        ;
    else if (piped_io_flag)
	out_str(tmpstr_bf);
    else
	dbuf_add(conb, tmpstr_bf, ++len, CONB_NEWLINE);
}

void conb_add_n(char *str)
{
    if (piped_io_flag)
	out_str_n(str);
    else
	dbuf_add_stri(conb, str, CONB_NO_NEWLINE);
}

void show_user(const char *frmt, ...)
{
    int len;
    va_list ap;

    va_start(ap, frmt);
    len = vsnprintf(tmpstr_bf, BUFF_SIZE-1, frmt, ap);
    va_end(ap);

    if (len == -1)
        ;
    else if (have_tty() || piped_io_flag)
	out_str(tmpstr_bf);
    else
	dbuf_add(conb, tmpstr_bf, ++len, CONB_NEWLINE);
}

void show_user_n(char *str)
{
    if (have_tty() || piped_io_flag)
	out_str_n(str);
    else
	dbuf_add_stri(conb, str, CONB_NO_NEWLINE);
}

/************************************
 *         File buffering           *
 ************************************/

void /* user */ mess_comm()
{
    char *opts, *file;
    int len, sort;
    FILE *fd;

    file = NULL;
    sort = FALSE;
    opts = dbuf_get(args, 1);

    if (opts && strcmp(opts, "-s") == 0) {
	sort = TRUE;
	file = dbuf_get(args, 2);
    } else
	file = opts;
    
    if (!file)
	print_cmnd_help(MESS_CMND);
    else if (!(fd = myfopen(file, "r")))
	show_user(DEBUG_MSG " Error opening \"%s\" : %s", file, 
		  strerror(errno));
    else {
        len = get_fline(fd, tmpstr_bf, BUFF_SIZE);
	while (!feof(fd)) {
	    dbuf_add(outb, tmpstr_bf, ++len, 0);
	    len = get_fline(fd, tmpstr_bf, BUFF_SIZE);
	}
	fclose(fd);

	if (sort) 
	    outb_sort();
	outb_flush();
    }
}

/************************************
 *      Decoder Log functions       *
 ************************************/

void /* user */ dlog_clear_comm()
{
    if (!dlog->len)
	show_user(DEBUG_MSG " %s is empty", dlog->name);
    else {
	dbuf_clear(dlog);
	show_user(STATUS_MSG " %s cleared", dlog->name);
    }
}

void /* user */ dlog_show_comm()
{
    int i;

    if (!dlog->len)
	outb_add(DEBUG_MSG " %s is empty", dlog->name);
    else
	for (i = 0; i < dlog->len; i++)
	    outb_add("%s", dbuf_get(dlog, i));

    outb_flush();
}

void dlog_add(const char *frmt, ...)
{
    int len;
    time_t tim;
    char *time_ascii;
    va_list ap;

    va_start(ap, frmt);
    len = vsnprintf(tmpstr_bf, BUFF_SIZE-1, frmt, ap);
    va_end(ap);

    if (len == -1)
        ;
    else {
	dbuf_add(dlog, tmpstr_bf, ++len, 0);
	tim = time(NULL);
	time_ascii = ctime(&tim);
	remove_newline(time_ascii);
	len = safe_sprintf(tmpstr_bf, BUFF_SIZE, "\t...logged %s", time_ascii);
	if (len != -1)
	    dbuf_add(dlog, tmpstr_bf, ++len, 0);
    }
}
