/***********************************************
 *  event.c
 *
 *  Copyright 2001 Eric Kratzer.
 *  Includes Xaudio software Copyright (C)
 *           1996-2000 Xaudio Corporation
 *           All Rights Reserved.
 *
 *  This file is part of imp3sh.
 *
 *  imp3sh is free software; you can distribute
 *  it and/or modify it under the terms of the
 *  GNU General Public License as published by
 *  the Free Software Foundation; either version
 *  2 of the License, or (at your option) any
 *  later version.
 *
 *  imp3sh is distributed in the hope that
 *  it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *  PURPOSE. See the GNU General Public License
 *  for more details.
 *
 *  You should have recieved a copy of the GNU
 *  General Public License along with imp3sh; if
 *  not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston MA
 *  02111-1307 USA
 ************************************************/

/************************************
 *          Include Files           *
 ************************************/

#include "imp3sh.h"
#include "all.h"

/************************************
 *       External References        *
 ************************************/

/* command.c */
extern CommandDef com[];
extern DCharBuf *args;

/* list.c */
extern ListDef *active_list;
extern char *current, *last_stream;

/* shell.c */
extern flag_t ignore_hist_input;

/************************************
 *        Global Variables          *
 ************************************/

char *YesNo[] = { 
    "No", 
    "Yes" 
};

/************************************
 *        Local Variables           *
 ************************************/

static char *EVENTYPES[] = { 
    "TMD (timed)",
    "EOL (end of list)",
    "EOQ (end of queue)",
    "EOS (end of stream)",
    "SOS (start of stream)",
    "EXT (exit)"
};
static char *EVENTIDS[] = { 
    "TMD", 
    "EOL", 
    "EOQ", 
    "EOS", 
    "S0S", 
    "EXT" 
};
static char tmpstr_ev[COMM_SIZE];
static unsigned int event_mask;
static int event_ids;
static EventData *ev_start, *ev_end;

/************************************
 *         Module routines          *
 ************************************/

unsigned int event_get_bytec()
{
    EventData *curr;
    unsigned int bytec;

    curr = ev_start;
    bytec = 0;

    while (curr) {
	bytec += sizeof(EventData);
	bytec += get_strlen(curr->command) + 1;
	bytec += get_strlen(curr->info) + 1;
	curr = curr->next;
    }

    return bytec;
}

void event_init()
{
    event_ids = event_mask = 0;
    ev_start = ev_end = NULL;
    memset(tmpstr_ev,  0, COMM_SIZE);
}

void event_clear()
{
    EventData *curr;

    curr = NULL;

    while (ev_start) {
	curr = ev_start->next;
	free(ev_start->command);
	free(ev_start->info);
	free(ev_start);
	ev_start = curr;
    }

    ev_start = ev_end = NULL;
    event_ids = 0;
}

/************************************
 *         Event Manipulation       *
 ************************************/

void event_post(unsigned int type)
{
    event_mask |= type;
}

void event_add(flag_t persist, flag_t quiet, unsigned int type,
	       struct timeval period_s, char *command, char *info)
{
    if (ev_start) {
	ev_end->next = malloc(sizeof(EventData));
	ev_end = ev_end->next;
    } else {
	ev_start = malloc(sizeof(EventData));
	ev_end = ev_start;
    }

    if (type == TIME_EVENT) {
	gettimeofday(&ev_end->start, NULL);
	ev_end->period = period_s;
    } else {
	memset(&ev_end->start,  0, sizeof(struct timeval));
	memset(&ev_end->period, 0, sizeof(struct timeval));
    }

    if (info)
	ev_end->info = strdup(info);
    else
	ev_end->info = NULL;

    ev_end->id = event_ids++;
    ev_end->type = type;
    ev_end->persist = persist;
    ev_end->quiet = quiet;
    ev_end->command = strdup(command);
    ev_end->next = NULL;
}

int event_remove(int event_id)
{
    EventData *curr, *prev;

    curr = ev_start;
    prev = NULL;

    while (curr) {
	if (curr->id == event_id)
	    break;
	else {
	    prev = curr;
	    curr = curr->next;
	}
    }

    if (curr) {
	if (curr == ev_end)
	    ev_end = prev;

	if (prev)
	    prev->next = curr->next;
	else
	    ev_start = curr->next;

	free(curr->command);
	free(curr->info);
	free(curr);

	if (ev_start == NULL)
	    event_ids = 0;

	return TRUE;
    }

    return FALSE;
}

char *event_get_str(unsigned int type)
{
    switch (type) {
    case TIME_EVENT:
	return EVENTYPES[0];
    case LIST_EVENT:
	return EVENTYPES[1];
    case QUEUE_EVENT:
	return EVENTYPES[2];
    case ESTREAM_EVENT:
	return EVENTYPES[3];
    case SSTREAM_EVENT:
	return EVENTYPES[4];
    case EXIT_EVENT:
	return EVENTYPES[5];
    default:
	return "Unknown event";
    }
}

char *event_get_id(unsigned int type)
{
    switch (type) {
    case TIME_EVENT:
	return EVENTIDS[0];
    case LIST_EVENT:
	return EVENTIDS[1];
    case QUEUE_EVENT:
	return EVENTIDS[2];
    case ESTREAM_EVENT:
	return EVENTIDS[3];
    case SSTREAM_EVENT:
	return EVENTIDS[4];
    case EXIT_EVENT:
	return EVENTIDS[5];
    default:
	return "Unknown";
    }
}

void event_run(EventData * ev)
{
    flag_t quiet, persist;

    quiet = ev->quiet;
    persist = ev->persist;

    if (persist)
	gettimeofday(&ev->start, NULL);

    safe_strcpy(tmpstr_ev, ev->command, COMM_SIZE);

    if (!quiet) {
	if (have_tty()) {
	    com_ctrl_str(1, TC_CR);
	    com_ctrl_str(1, TC_CLR_TOEOL);
	    out_flush();
	}
	show_user(STATUS_MSG " Executing %s event [%d]:",
		  event_get_id(ev->type), ev->id);
    }

    ignore_hist_input = TRUE;
    com_handle(tmpstr_ev);
    ignore_hist_input = FALSE;

    if (!quiet)
	show_user(STATUS_MSG " Event [%d] done.", ev->id);

    if (!persist)
	event_remove(ev->id);

    if (!quiet && have_tty())
	update_prompt(FALSE);
}

void event_check_list(EventData * curr)
{
    if (!curr->info)
	return;

    if (strcmp(curr->info, "*") == 0 || 
	(active_list && strcmp(curr->info, active_list->dcb->name) == 0))
	event_run(curr);
}

void event_check_stream(EventData * curr, char *str)
{
    int len, len2;

    if (!curr->info)
	return;

    if (strcmp(curr->info, "*") == 0)
	event_run(curr);
    else {
	len = get_strlen(curr->info);
	len2 = get_strlen(str);
	if (len && len2 && len2 > len &&
	    strcasecmp(&str[len2 - len], curr->info) == 0)
	    event_run(curr);
    }
}

int event_check_time(struct timeval *st, struct timeval *pe)
{
    struct timeval tnow;
    unsigned long sec, usec;

    gettimeofday(&tnow, NULL);

    sec = st->tv_sec + pe->tv_sec;
    usec = st->tv_usec + pe->tv_usec;

    if (usec >= 1000000) {
	sec ++;
	usec -= 1000000;
    }

    if (tnow.tv_sec > sec || (tnow.tv_sec == sec && tnow.tv_usec >= usec))
	return TRUE;
	  
    return FALSE;
}

void event_check()
{
    EventData *temp, *curr;
    unsigned int mask;

    temp = NULL;
    curr = ev_start;
    mask = event_mask;
    event_mask = 0;

    while (curr) {
	temp = curr->next;

	switch (curr->type) {
	case TIME_EVENT:
	    if (event_check_time(&curr->start, &curr->period))
		event_run(curr);
	    break;
	case LIST_EVENT:
	    if (mask & LIST_EVENT)
		event_check_list(curr);
	    break;
	case QUEUE_EVENT:
	    if (mask & QUEUE_EVENT)
		event_run(curr);
	    break;
	case ESTREAM_EVENT:
	    if (mask & ESTREAM_EVENT)
		event_check_stream(curr, last_stream);
	    break;
	case SSTREAM_EVENT:
	    if (mask & SSTREAM_EVENT)
		event_check_stream(curr, current);
	    break;
	case EXIT_EVENT:
	    if (mask & EXIT_EVENT)
		event_run(curr);
	    break;
	}

	curr = temp;
    }
}

void event_print_add_help()
{
    print_cmnd_help(ADD_EVENT);

    show_user(STATUS_MSG " Use '-p' to make event persistent (repeating)");
    show_user(STATUS_MSG " Use '-d' to disable event alerts (quiet)");
    show_user(STATUS_MSG
	      " For <type_options...>, use one of the following:");
    show_user("\t1) -t <seconds>           : TMD (timed) event");
    show_user("\t2) -l ('*' | <list_name>) : EOL (end of list) event");
    show_user("\t3) -q                     : EOQ (end of queue) event");
    show_user("\t4) -s ('*' | <filetype>)  : EOS (end of stream) event");
    show_user("\t5) -r ('*' | <filetype>)  : SOS (start of stream) event");
    show_user("\t6) -e                     : EXT (exit) event");
    show_user(STATUS_MSG
	      " For <command>, enter command(s) as you would by themselves");
}

void /* user */ event_add_comm()
{
    char *opt, *time, *info;
    int cm_ndx;
    double time_f;
    struct timeval time_s;
    unsigned int type;
    flag_t persist, quiet;

    cm_ndx = 1;
    opt = dbuf_get(args, cm_ndx++);
    time = info = NULL;
    type = TIME_EVENT;
    persist = quiet = FALSE;

    if (!opt) {
	event_print_add_help();
	return;
    }

    while (opt) {
	if (strcmp(opt, "-p") == 0)
	    persist = TRUE;
	else if (strcmp(opt, "-d") == 0)
	    quiet = TRUE;
	else if (strcmp(opt, "-e") == 0) {
	    type = EXIT_EVENT;
	    break;
	} else if (strcmp(opt, "-t") == 0) {
	    type = TIME_EVENT;
	    time = dbuf_get(args, cm_ndx++);
	    if (time) {
		time_f = atof(time);
		if (time_f == 0.0) {
		    show_user(DEBUG_MSG 
			      " Invalid length for timed event: \"%s\"", 
			      time);
		    return;
		}
		time_s.tv_sec = (long)time_f;
		time_s.tv_usec = (long)(time_f*1000000.0 - 
					(double)time_s.tv_sec*1000000.0);
	    } else {
		event_print_add_help();
		return;
	    }
	    break;
	} else if (strcmp(opt, "-l") == 0) {
	    type = LIST_EVENT;
	    info = dbuf_get(args, cm_ndx++);

	    if (!info) {
		event_print_add_help();
		return;
	    }

	    break;
	} else if (strcmp(opt, "-q") == 0) {
	    type = QUEUE_EVENT;
	    break;
	} else if (strcmp(opt, "-s") == 0) {
	    type = ESTREAM_EVENT;
	    info = dbuf_get(args, cm_ndx++);

	    if (!info) {
		event_print_add_help();
		return;
	    }

	    break;
	} else if (strcmp(opt, "-r") == 0) {
	    type = SSTREAM_EVENT;
	    info = dbuf_get(args, cm_ndx++);

	    if (!info) {
		event_print_add_help();
		return;
	    }

	    break;
	} else {
	    event_print_add_help();
	    return;
	}
	opt = dbuf_get(args, cm_ndx++);
    }

    *tmpstr_ev = ZCHAR;
    opt = dbuf_get(args, cm_ndx++);
    while (opt) {
	strcat(tmpstr_ev, opt);
	opt = dbuf_get(args, cm_ndx++);
	if (opt)
	    strcat(tmpstr_ev, " ");
    }

    if (*tmpstr_ev == ZCHAR)
	event_print_add_help();
    else
	event_add(persist, quiet, type, time_s, tmpstr_ev, info);
}

void /* user */ event_remove_comm()
{
    char *id_str;
    int id;

    id_str = dbuf_get(args, 1);

    if (!id_str)
	print_cmnd_help(REMOVE_EVENT);
    else if (!ev_start)
	show_user(STATUS_MSG " Event list is empty");
    else if (is_all_digits(id_str)) {
	id = atoi(id_str);
	if (!event_remove(id))
	    show_user(DEBUG_MSG " Event ID %d not found", id);
    } else if (strcmp(id_str, "all") == 0) {
	event_clear();
	show_user(STATUS_MSG " Event list cleared");
    } else
	print_cmnd_help(REMOVE_EVENT);
}

void /* user */ event_print_comm()
{
    EventData *curr;
    struct timeval now, next;
    char tmpt[64];

    curr = ev_start;
    gettimeofday(&now, NULL);

    if (!curr) {
	show_user(DEBUG_MSG " Event list is empty");
	return;
    }

    while (curr) {
	outb_add(MISC_MSG " Event ID [%d]", curr->id);
	outb_add("\tType       : %s", event_get_str(curr->type));

	if (curr->info)
	    switch (curr->type) {
	    case LIST_EVENT:
		if (strcmp(curr->info, "*") == 0)
		    outb_add("\tList       : (any)");
		else {
		    outb_add("\tList       : \"%s\"", curr->info);
		    if (!plist_get(curr->info))
			outb_add("\tWARNING    : \"%s\" does not exist",
				 curr->info);
		}
		break;
	    case ESTREAM_EVENT:
	    case SSTREAM_EVENT:
		if (strcmp(curr->info, "*") == 0)
		    outb_add("\tFiletype   : (any)");
		else
		    outb_add("\tFiletype   : %s", curr->info);
		break;
	    }

	outb_add("\tCommand    : %s", curr->command);
	if (curr->type != EXIT_EVENT)
	    outb_add("\tPersistent : %s", YesNo[curr->persist]);
	outb_add("\tQuiet      : %s", YesNo[curr->quiet]);

	if (curr->type == TIME_EVENT) {
	    next = add_timevals(curr->start, curr->period);
	    if (curr->persist) {
		build_msec_time_string(curr->period, tmpt, 64);
		outb_add("\tPeriod     : %s", tmpt);
	    }
	    build_msec_time_string(sub_timevals(next, now), tmpt, 64);
	    outb_add("\tNext       : %s", tmpt);
	} else if (curr->start.tv_sec > 0) {
	    build_msec_time_string(sub_timevals(now, curr->start), tmpt, 64);
	    outb_add("\tLast       : -%s", tmpt);
	}
	curr = curr->next;
    }

    outb_flush();
}
